# SCRIPT FOR SENDING PUSHOVER NOTIFICATIONS WITH A CUSTOM RESIZED ALERT IMAGE ATTACHMENT
#
# Version: 2.0
# Created by jaydeel at ipcamtalk.com
# https://ipcamtalk.com/threads/send-pushover-notifications-with-pictures-and-hyperlinks.58819/post-769688
#
# Updates:
# V2.0  - 2024-09-08 - BREAKING CHANGE: must replace old user-setting file 
#                    - BREAKING CHANGE: eliminated support for positional arguments; now must use named arguments
#                    - BREAKING CHANGE: converted arguments -Quiet and -Demo from Y/N strings to switches
#                    - BREAKING CHANGE: default path for resized images is no longer the Blue Iris Alerts folder (we no 
#                      longer add files to Blue Iris-manged folders); please provide a path in user setting $img_path
#                    - FIXED:: fixed bug in function SendPushover()
#                    - ADDED: arguments -Sound, -Devices to override global settings in the user-settings file
#                    - ADDED: argument -Method to override the script's default method when using argument -TargetKB
#                    - ADDED: argument -Help to display usage guide (also displayed when executing script with no arguments)
#                    - IMPROVEMENT: script's console output categorized and clarified
# V1.9  - 2023-07-12 - FIXED: bug with Pushover -Priority parameter (mis-typed as [String] instead of [Int])
#                    - REWORKED: function 'PurgeOldFile()'
# V1.8  - 2023-06-04 - ADDED: options to tune image resizing: 'WxH' (default), 'Quality' and 'Hybrid'; 
#                      see USER SETTINGS $targetkb_method
#                    - ADDED: new named argument '-Demo Y' to suspend notifications & BI logfile posting 
#                      (useful for tuning resize settings)
#                    - ADDED: automatic renaming of user settings file; changed hard-coded filename, 
#                      now using 'THIS_SCRIPTNAME_user_settings.ps1'
#                    - FIXED: bugs for arguments -Quality, -Width & -TargetKB (conditionals incorrectly 
#                      assumed [Int] typing)
#                    - FIXED: bugs that prevented the posting of (some) error messages to the Blue Iris logfile
# V1.7  - 2023-05-31 - FIXED: bugs & clarified user settings comments for posting messages to the Blue Iris logfile
# V1.6  - 2023-05-30 - MOVED: USER SETTINGS code to a separate configuration file (pushover_resize_user_settings.ps1)
#                    - ADDED: named argument '-Title' (for custom notification titles)
#                    - ADDED: 5 user settings to support posting error messages to the Blue Iris logfile
#                      (use setting '$post_to_BI_logfile_flg' to enable/disable this feature)
#                    - ADDED: a user setting to post ALL resize event messages to the Blue Iris logfile ($post_all_events)
#                    - IMPROVED: background messaging toggling
# V1.5  - 2023-05-28 - ADDED: 7 named commandline arguments: -Msg, -Priority, -AttachFile, -Quality, -Width, -TargetKB -Quiet
#                      using any of the last five will override the defaults in the USER SETTINGS section
#                    - CHANGED: user setting '$img_size' to '$img_width' (for clarity of its function)
#                    - FIXED: potential memory leak opportunity in function ResizeImage()
#                    - ADDED: a background message that summarizes attachment resize results and script performance
#                    - ADDED: WxHBefore and WxHAfter columns to CSV log (must rename existing CSV file before they will appear)
# V1.4  - 2023-05-27 - CHANGED: attachment argument requirement to optional; script now requires only message argument
# V1.3  - 2023-05-26 - ADDED: Powershell's execution policy info to the background message
#                    - ADDED: alert jpg's filename to the background message
#                    - ADDED: '-UseQuotes AsNeeded' to 'Export-Csv' cmdlet in function LogScriptResultsToCsv()
# V1.2  - 2023-05-25 - ADDED: user setting for background messaging
# V1.1  - 2023-05-24 - FIXED: bug in function ResizeImage()
# V1    - 2023-05-24 - initial release
#
# USAGE:      {this script name}.ps1 Msg [Priority AttachFile Quality Width, TargetKB Quiet]
# ARGUMENTS:  see param lines below
#
# EXAMPLES:
# 	Send notification without an attachment
#   	pushover_resize.ps1 'test msg'
#   	pushover_resize.ps1 -Msg 'test msg' -Title "test title"
# 	Send notification without an attachment with custom title and priority
#   	pushover_resize.ps1 -Msg 'test msg' -Title "test title" -Priority -1
# 	Send notification with default priority and with attachment as-s (not resized)
#   	pushover_resize.ps1 'test msg' * 'D:\BlueIris\Alerts\DW1.20230521_160000.3076088.17-0.jpg' 100 0
#   	pushover_resize.ps1 -Msg 'test msg' -Title "test title" -AttachFile 'D:\BlueIris\Alerts\DW1*.jpg' -Quality 100 -Width 0
# 	Change only img_quality, use default img_width
#   	pushover_resize.ps1 'test msg' * 'D:\BlueIris\Alerts\DW1.20230521_160000.3076088.17-0.jpg' 70
# 	Change only img_width, use default img_quality
#		pushover_resize.ps1 'test msg' * 'D:\BlueIris\Alerts\DW1.20230521_160000.3076088.17-0.jpg' * 500
#		pushover_resize.ps1 -Msg 'test msg' 'D:\BlueIris\Alerts\DW1.20230521_160000.3076088.17-0.jpg' -Width 500
# 	Change both img_quality and img_width
#		pushover_resize.ps1 'test msg' * 'D:\BlueIris\Alerts\DW1.20230521_160000.3076088.17-0.jpg' 70 500
# 	Target attachment size of 500 KB, using default img_quality & img_width
#		pushover_resize.ps1 'test msg' * 'D:\BlueIris\Alerts\DW1.20230521_160000.3076088.17-0.jpg' * * 500
#		pushover_resize.ps1 -Msg 'test msg' -Priority * -AttachFile 'D:\BlueIris\Alerts\DW1*.jpg' -TargetKB 500
# 	Target attachment size of 500 KB, using specified img_quality
#		pushover_resize.ps1 'test msg' * 'D:\BlueIris\Alerts\DW1.20230521_160000.3076088.17-0.jpg' 70 * 500
#		pushover_resize.ps1 -Msg 'test msg' -AttachFile 'D:\BlueIris\Alerts\DW1.20230521_160000.3076088.17-0.jpg' -Quality -TargetKB 500

# MUST BE THE FIRST LINE IN A SCRIPT
param(
	[String]$Msg,         # [REQUIRED] Pushover notification message - may contain HTML markup
	[Int]$Priority,       # [OPTIONAL] -2|-1|0|1|2; -2=lowest, -1=low(silent), 0=normal (default), 1=high-priority, 2=emergency
	[String]$AttachFile,  # [OPTIONAL] image full filespec (typ. in ALERTS folder), typ. via Blue Iris macro &ATTACH_PATH
	[Int]$Quality,        # [OPTIONAL] target image quality 1-100; units %
	[Int]$Width,          # [OPTIONAL] target image size, smallest axis; units pixels
	[Int]$TargetKB,       # [OPTIONAL] target filesize in kilobytes; resized file will be <= this value
	[String]$Method,      # [OPTIONAL] custom method for argument -TargetKB: 'WxH' (default), 'Quality' or 'Hybrid'
	[String]$Title,       # [OPTIONAL] custom title for pushover message
	[String]$Sound,       # [OPTIONAL] custom sound for pushover message
	[String]$Devices,     # [OPTIONAL] custom devices for pushover message, comma-delimited
	[Switch]$Quiet,       # [OPTIONAL] disables background messaging
	[Switch]$Demo,        # [OPTIONAL] disables sending notifications, and posting to the Blue logfile
	[Switch]$Help         # [OPTIONAL] show usage guide
)

# For debugging
#Write-Host "Msg = $Msg"
#Write-Host "Priority = $Priority"
#Write-Host "AttachFile = $AttachFile"
#Write-Host "Quality = $Quality"
#Write-Host "Width = $Width"
#Write-Host "TargetKB = $TargetKB"
#Write-Host "Method = $Method"
#Write-Host "Title = $Title"
#Write-Host "Sound = $Sound"
#Write-Host "Devices = $Devices"
#Write-Host "Quiet = $Quiet"
#Write-Host "Demo = $Demo"
#Write-Host "Help = $Help"

#Set-StrictMode  -Version 3
#Start-Transcript  # default destination is C:\Users\{user}l\Documents


# GATHER/VALIDATE SCRIPT INFO 
if (1 -eq 1) {  # Use to collapse this code section while editing

	#$ScriptName = $MyInvocation.MyCommand.Name
	$ScriptName = Split-Path -leaf $PSCommandpath
	$ScriptPath = $PSCommandPath.Replace("\$ScriptName", '')
	$ScriptBaseName = $ScriptName.Replace('.ps1', '')
	$ScriptPath = $PSCommandPath.Replace("\$ScriptName", '')

	# For debugging
	# Write-Host "ScriptName = $ScriptName"
	# Write-Host "ScriptPath = $ScriptPath"
	# Write-Host "ScriptBaseName = $ScriptBaseName"
	# Write-Host "ScriptPath = $ScriptPath"

	# INITIALIZE VARS
	$PsVersion = $PSVersionTable.PSVersion.Major
	# NOTE: Scripts running from Windows Task Scheduler have a parent process named 'svchost'.
	#       https://stackoverflow.com/a/65894924
	$ParentProcess = (Get-Process -Id (Get-CimInstance Win32_Process -Filter "ProcessID = $pid").ParentProcessId).Name
	# NOTE: Scripts called by another script have a 'PSCommandPath'.
	#       Method 'PSCommandPath' returns null if script is executed via the terminal
	#       https://stackoverflow.com/a/23370618/13499984
	$CallingScript = $MyInvocation.PSCommandPath

	# DETERMINE IF RUNNING IN BACKGROUND (IN CLI MODE, OR FROM BLUE IRIS)
	if ($ParentProcess -eq 'BlueIris') {
		$FromConsole = $false
	} else {
		$FromConsole = ($null -eq $CallingScript)
	}

	# For debugging
	#msg.exe * /TIME:10 "entered script '$ScriptName'"
	#msg.exe * /TIME:10 "PsVersion ($ScriptName) = $PsVersion"
	#msg.exe * /TIME:10 "ParentProcess ($ScriptName) = $ParentProcess"
	#msg.exe * /TIME:10 "fromConsole ($ScriptName) = $FromConsole"

	# SHOW OPTIONAL MESSAGE WHEN RUNNING FROM BACKGROUND
	if (!$FromConsole) {
		if ($bg_msgs_flg ) {
			$ExecPolicy = Get-ExecutionPolicy
			$exec_policy_str = "PSExecutionPolicy = $ExecPolicy"
			if ($env:PSExecutionPolicyPreference) {
				$ExecPolicyPref = $env:PSExecutionPolicyPreference
				$exec_policy_str = "Session ExecutionPolicy = $ExecPolicyPref"
			}
			$msg_sec = 10
			$bg_msg = "Script '$ScriptName' is running in the background" + 
			"`n$exec_policy_str" + 
			"`nPsVersion = $PsVersion" + 
			"`nParentProcess = $ParentProcess" + 
			"`nalert = $AttachFile" + 
			"`n`n(this msg auto-closes in $msg_sec seconds)"
			msg.exe * /TIME:$msg_sec $bg_msg
		}
	}

	# For testing CLI mode
	#$FromConsole = $false

	# CHECK IF SCRIPT'S PATH IS IN ENVIRONMENT 'SYSTEM PATH' VARIABLE
	<# $EnvPath = [Environment]::GetEnvironmentVariable('Path', 'Machine')
	if ($null -ne $EnvPath) {
		if (!($EnvPath -split ';' -contains $ScriptPath)) {

			# For debugging
			#Write-Host "ScriptPath = $ScriptPath"
			#Write-Host ("EnvPath = " + (($EnvPath -Split ';') | ConvertTo-Json))

			return "ERR/this script's path '$ScriptPath' does not exist in the Environment System Path variable"
			
			# Attempt repair
			# NOTE: when I tried this it added MANY other/duplicate entries
			#[Environment]::SetEnvironmentVariable("Path", $EnvPath + ";$ScriptPath", "Machine")
			#Write-Warning "Added this script's path '$ScriptPath' to the Environment Path"
		}
	} else {
		Write-Warning "Unable to inspect Environment variable; be sure script's path is in the Environment Path"
	} #>
	# For debugging
	#Write-Host ("EnvPath = " + (($EnvPath -Split ';') | ConvertTo-Json))

	# For testing - Simulate CLI mode
	#$FromConsole = $false

	# FEEDBACK UNEXPECTED COMMANDLINE ARGUMENTS
	#Write-Host ("params = " + ($PSBoundParameters | ConvertTo-Json))
	$cmdline_args_passed = [String[]]$PSBoundParameters.Keys
	#Write-Host "Extra Args w/ values = $Args"
	$cmdline_args_bad = ($Args | where{$_ -like "-*"}) | ForEach-Object{$_ -Replace '-', ''}
	if ($cmdline_args_bad) {
		$cmdline_args_bad = [String[]]$cmdline_args_bad
	} else {
		$cmdline_args_bad = @()
	}

	# For debugging
	#Write-Host "Cmdline Args Passed = $cmdline_args_passed"
	#Write-Host "Cmdline Args Bad = $cmdline_args_bad"

	if ($cmdline_args_bad.Count -gt 0) {
		if ($FromConsole) {
			Write-Host ("HEADS-UP: commandline contains unsupported argument(s): [ " + ($cmdline_args_bad -Join ', ') + " ]") -ForegroundColor Green
			return
		}
	}

}

# ================================================
# IMPORT USER SETTINGS...

# From external 'user settings' script file (expected to be in same folder as this script)
$user_settings_filename = $ScriptName.Replace(".ps1", "_user_settings.ps1")
try {
	. $user_settings_filename
	
	# For debugging
	#Write-Host "Loaded user settings from '$user_settings_filename'"
}
catch {
	return "ERR/failed to load user settings file '$user_settings_filename'`nTip: try executing it directly to confirm it exists and reveal syntax errors"
}

# END USER SETTINGS
# ===============================================

# Parse user settings
if (1 -eq 1) {  # Use to collapse this code section while editing

	$variable_test = Get-Variable -Name bg_msgs_flg -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$bg_msgs_flg = $true
	}
	$variable_test = Get-Variable -Name targetkb_method -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$targetkb_method = 'WxH'
	}
	if (@('WxH','Quality','Hybrid') -notcontains $targetkb_method) {
		$targetkb_method = 'WxH'
	}
	$variable_test = Get-Variable -Name targetkb_wxh_step -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$targetkb_wxh_step = 25
	}
	$variable_test = Get-Variable -Name targetkb_quality_step -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$targetkb_quality_step = 20
	}
	$variable_test = Get-Variable -Name img_quality -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$img_quality = 85
	}
	$variable_test = Get-Variable -Name img_width -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$img_width = 500
	}
	$variable_test = Get-Variable -Name img_kilobytes -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$img_kilobytes = 0
	}

	$variable_test = Get-Variable -Name img_path -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$err_msg = "ERR/user-settings file is missing setting `$img_path"
		return $err_msg
	} else {
		if ($img_path.StartsWith('YOUR')) {
			$err_msg = "ERR/user-setting file setting `$img_path is not set"
			return $err_msg
		}
		if (!$img_path.EndsWith('\')) {
			$img_path = $img_path + '\'
		}
	}

	$variable_test = Get-Variable -Name img_name_suffix -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$img_name_suffix = 'resized'
	}

	$variable_test = Get-Variable -Name po_app_token -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$err_msg = "ERR/user-settings file is missing setting `$po_app_token"
		return $err_msg
	} else {
		if ($po_app_token.StartsWith('YOUR')) {
			$err_msg = "ERR/user-setting file setting `$po_app_token is not set"
			return $err_msg
		}
	}
	$variable_test = Get-Variable -Name po_user_token -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$err_msg = "ERR/user-settings file is missing setting `$po_user_token"
		return $err_msg
	} else {
		if ($po_user_token.StartsWith('YOUR')) {
			$err_msg = "ERR/user-setting file setting `$po_user_token' is not set"
			return $err_msg
		}
	}

	$variable_test = Get-Variable -Name img_lifetime_hours -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$img_lifetime_hours = 24
	}

	$variable_test = Get-Variable -Name po_priority -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$po_priority = 0
	}
	$variable_test = Get-Variable -Name po_sound -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$po_sound = ''
	}
	$variable_test = Get-Variable -Name po_device -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$po_device = ''
	}
	$variable_test = Get-Variable -Name po_retry -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$po_retry = 30
	}
	$variable_test = Get-Variable -Name po_expire -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$po_expire = 300
	}
	$variable_test = Get-Variable -Name po_tags -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$po_tags = ''
	}
	
	$variable_test = Get-Variable -Name post_to_BI_logfile_flg -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$post_to_BI_logfile_flg = $false
	}
	$variable_test = Get-Variable -Name post_all_events -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$post_all_events = $false
	}
	$variable_test = Get-Variable -Name post_w_creds_flg -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$post_w_creds_flg = $false
	}
	if ($post_to_BI_logfile_flg) {
		$variable_test = Get-Variable -Name bi_webserver -Scope Local -ErrorAction SilentlyContinue
		if ($null -eq $variable_test) {
			$err_msg = "ERR/user-settings file is missing setting `$bi_webserver"
			return $err_msg
		} else {
			if ($bi_webserver.StartsWith('YOUR')) {
				$err_msg = "ERR/user-setting file setting '`$bi_webserver' is not set"
				return $err_msg
			}
		}
		if ($post_w_creds_flg) {
			$variable_test = Get-Variable -Name bi_user_name -Scope Local -ErrorAction SilentlyContinue
			if ($null -eq $variable_test) {
				$err_msg = "ERR/user-settings file is missing setting `$bi_user_name"
				return $err_msg
			} else {
				if ($bi_user_name.StartsWith('YOUR')) {
					$err_msg = "ERR/user-setting file setting '`$bi_user_name' is not set"
					return $err_msg
				}
			}
			$variable_test = Get-Variable -Name bi_user_pw -Scope Local -ErrorAction SilentlyContinue
			if ($null -eq $variable_test) {
				$err_msg = "ERR/user-settings file is missing setting `$bi_user_pw"
				return $err_msg
			} else {
				if ($bi_user_pw.StartsWith('YOUR')) {
					$err_msg = "ERR/user-setting file setting '`$bi_user_pw' is not set"
					return $err_msg
				}
			}
		}
	}
}

# ===============================================

function PostCustomBiLogEntry() {
# PURPOSE:  Add custom message to the Blue Iris logfile
# ARGS:     $MsgStr        = (string) may contain HTML
#           $Level         = (integer) 0|1|2 (0:info, 1:warn, 2:error)
#           $SilentFlg     = (boolean) $true prevents background messaging
# RETURNS:  string
# REQUIRES: none
# GLOBALS:  Uses vars: $script:bi_webserver, $script:post_w_creds_flg, $script:bi_user_name, $script:bi_user_pw
# KEY POINT: the message string maximum length is 128 characters (BI HTTP interface command '/admin?log=message&level=x')

    param(
		[String]$MsgStr,      # may contain HTML
		[String]$Level,       # full filespec of resized alert image
		[Boolean]$SilentFlg   # 
	)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	# Write-Host "Entered $fn_name() ..."
	# Write-Host "$fn_name() MsgStr = '$MsgStr"
	# Write-Host "$fn_name() Level = $Level"
	# Write-Host "$fn_name() SilentFlg = $SilentFlg"
	
	# Remove/replace incompatible substrings from $MsgStr
	$MsgStr = $MsgStr -Replace "`n", "%20" -Replace "\\r",  "%20" -Replace "\\n", "%20" -Replace " ", "%20"
	
	$bi_address = $script:bi_webserver
	if ($script:post_w_creds_flg) {
		$uri = "http://$bi_address/admin?log=$MsgStr&level=$Level&user=$script:bi_user_name&pw=$script:bi_user_pw"
	} else {
		$uri = "http://$bi_address/admin?log=$MsgStr&level=$Level"
	}
	$result = Invoke-WebRequest -Uri $uri
	
	# For debugging
	# Write-Host "$fn_name() uri = $uri"
	# Write-Host "$fn_name() result:`n$result"	

	# HANDLE NOTIFICATIONS
	if (!$SilentFlg) {
		msg.exe * /TIME:10 "Custom BI logfile message posted...`nlevel = $Level`n'$MsgStr'"
	}
	
	$rtn_str = "Custom BI logfile message posted:`n'" + ($MsgStr.Replace("%20"," ")) + "'"
	
	return $rtn_str
}

function ResizeImage() {
# PURPOSE:  Resize image file to be used as attachment for Pushover notification 
# ARGS:     see param lines below
# RETURNS:  (array) @(before_KB, after_KB, before_WH, after_WH)
# NOTES:    Be sure to check if 'after_KB' = 0, as this indicates the new file was not actually created
# REQUIRES: Functions: PostCustomBiLogEntry()
# GLOBALS: Uses vars: $script:FromConsole, $script:post_to_BI_logfile_flg, $script:post_all_events
#                     $script:ScriptName
# ATTRIBUTIONS:
#   Derived from: https://www.lewisroberts.com/2015/01/18/powershell-image-resize-function/
#   Also used: https://gist.github.com/someshinyobject/617bf00556bc43af87cd
	
    param(
		[String]$SourceFilespec,          # full filespec
		[String]$NewImagePath,            # folder location for resized files
		[Int]$Quality = 85,               # integer 0 to 100 (Blue Iris default is 85%)
		[Int]$Width = 500,                # pixels
		[Int]$TargetKb,                   # kilobytes; NOTE: Pushover API max is 2,621,440 bytes (2621 KB)
		[String]$TargetKbMethod = 'WxH',  # 'WxH', 'Quality', or 'Hybrid'
		[Int]$TargetKbWxHDelta = 25,      # percentage to decrease WxH per iteration
		[Int]$TargetKbQualityDelta = 25,  # percentage to decrease Quality per iteration
		[Boolean]$SilentFlg               # $true prevents background messaging (showing file resizing results)
	)
	
	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	# Write-Host "Entered function = $fn_name()"	
	# Write-Host "$fn_name() SourceImagePath = '$SourceFilespec'"
	# Write-Host "$fn_name() NewImagePath = '$NewImagePath'"
	# Write-Host "$fn_name() Quality = $Quality"
	# Write-Host "$fn_name() Width = $Width"
	# Write-Host "$fn_name() TargetKb = $TargetKb"
	# Write-Host "$fn_name() TargetKbMethod = $TargetKbMethod"
	# Write-Host "$fn_name() TargetKbWxHDelta = $TargetKbWxHDelta"
	# Write-Host "$fn_name() TargetKbQualityDelta = $TargetKbQualityDelta"
	
	if (-not(Test-Path -LiteralPath "$SourceFilespec") ) {
		$oldSizeKb = $null
	}

	$OutputLocation = $NewImagePath
	
    Add-Type -AssemblyName "System.Drawing"
 
    $img = [System.Drawing.Image]::FromFile($SourceFilespec)
 
	$oldHeight = $img.Height
	$oldWidth  = $img.Width
	$oldSizeKb = [math]::Round((Get-Item -Path $SourceFilespec).Length/1kB,1).ToString("#.0")
	$oldWH = "$oldWidth x $oldHeight"
	
	if ($TargetKb -gt 0) {
		# SPECIAL CASE - ITERATE RESIZING UNTIL FILE SIZE IS BELOW TARGET KILOBYTES
					
		if ($script:FromConsole) {
			Write-Host "File TargetKB details:"
			Write-Host "----------------------"
			Write-Host "Iterating to achieve file size < $TargetKb KB (using method '$TargetKbMethod')..."
		}
		
		if ($TargetKbMethod -eq 'WxH') {
			$i = 1
			$step = $TargetKbWxHDelta
			for ($x=(100-$step); $x -gt 0; $x=$x-$step) {
				
				# Show progress
				if ($script:FromConsole) {
					Write-Host ("Trying size reduction = -" + (100-$x) + "%")
				}
				
				$k = $x/100
				$CanvasHeight = $oldHeight*$k
				$CanvasWidth  = $oldWidth*$k

				# Encoder parameter for image quality
				$ImageEncoder = [System.Drawing.Imaging.Encoder]::Quality
				$encoderParams = New-Object System.Drawing.Imaging.EncoderParameters(1)
				$encoderParams.Param[0] = New-Object System.Drawing.Imaging.EncoderParameter($ImageEncoder, $Quality)
			 
				# Get codec
				$Codec = [System.Drawing.Imaging.ImageCodecInfo]::GetImageEncoders() | Where {$_.MimeType -eq 'image/jpeg'}
			 
				# Compute the final ratio to use
				$ratioX = $CanvasWidth / $img.Width;
				$ratioY = $CanvasHeight / $img.Height;
			 
				$ratio = $ratioY  # Default
				if ($ratioX -le $ratioY) {
					$ratio = $ratioX
				}
			 
				$newWidth = [Int] ($img.Width * $ratio)
				$newHeight = [Int] ($img.Height * $ratio)

				$bmpResized = New-Object System.Drawing.Bitmap($newWidth, $newHeight)
				#$bmpResized = New-Object System.Drawing.Bitmap -ArgumentList $newWidth, $newHeight
				$graph = [System.Drawing.Graphics]::FromImage($bmpResized)
				$graph.InterpolationMode = [System.Drawing.Drawing2D.InterpolationMode]::HighQualityBicubic
			 
				$graph.Clear([System.Drawing.Color]::White)
				$graph.DrawImage($img, 0, 0, $newWidth, $newHeight)
			 
				# Save to file
				$bmpResized.Save($OutputLocation, $Codec, $($encoderParams))
				$bmpResized.Dispose()
			
				if (-not(Test-Path -LiteralPath "$NewImagePath") ) {
					$err_msg = "ERR/Resize failed; cannot find NewImagePath '$NewImagePath'"
					if ($script:post_to_BI_logfile_flg -and $script:post_all_events) {
						PostCustomBiLogEntry `
						-MsgStr "$err_msg ($script:ScriptName)" `
						-Level 2 `
						-SilentFlg $SilentFlg
					}
					if ($script:FromConsole) {
						Write-Host $err_msg
					}
					msg.exe * /TIME:10 $err_msg
					
					# Release resource to avoid memory leak
					$img.Dispose()  
					
					return @($oldSizeKb, 0, $oldWH, "")

				} else {
					$newSizeKb = [math]::Round((Get-Item -Path $NewImagePath).Length/1kB, 1)
					$newWH = "$newWidth x $newHeight"
				}
				
				# Show progress
				if ($script:FromConsole) {
					Write-Host "--- result #$i height/width/size(KB) = $newHeight/$newWidth/$newSizeKb"
				}

				if ($newSizeKb -lt [Int]$TargetKb) {
					if ($script:FromConsole) {
						Write-Host "Achieved target after $i iterations"
					}
					break				
				} else {
					if ($x -le $step) {
						if ($script:FromConsole) {
							Write-Host ("Unable to achieve target after $i iterations; using final try results")
						}
					}
				}	
				$i++
			}
		}
		
		if ($TargetKbMethod -eq 'Quality') {
			$i = 1
			$step = $TargetKbQualityDelta  # percentage to decrease WxH per iteration
			for ($q=($Quality-$step); $q -gt 0; $q=$q-$step) {
				
				# Show progress
				if ($script:FromConsole) {
					Write-Host ("Trying quality reduction = -" + ($i*$step) + "%")
				}

				# Encoder parameter for image quality
				$ImageEncoder = [System.Drawing.Imaging.Encoder]::Quality
				$encoderParams = New-Object System.Drawing.Imaging.EncoderParameters(1)
				$encoderParams.Param[0] = New-Object System.Drawing.Imaging.EncoderParameter($ImageEncoder, $q)
			 
				# Get codec
				$Codec = [System.Drawing.Imaging.ImageCodecInfo]::GetImageEncoders() | Where {$_.MimeType -eq 'image/jpeg'}
				$newWidth = $img.Width
				$newHeight = $img.Height

				$bmpResized = New-Object System.Drawing.Bitmap($newWidth, $newHeight)
				#$bmpResized = New-Object System.Drawing.Bitmap -ArgumentList $newWidth, $newHeight
				$graph = [System.Drawing.Graphics]::FromImage($bmpResized)
				$graph.InterpolationMode = [System.Drawing.Drawing2D.InterpolationMode]::HighQualityBicubic
			 
				$graph.Clear([System.Drawing.Color]::White)
				$graph.DrawImage($img, 0, 0, $newWidth, $newHeight)
			 
				# Save to file
				$bmpResized.Save($OutputLocation, $Codec, $($encoderParams))
				$bmpResized.Dispose()
			
				if (-not(Test-Path -LiteralPath "$NewImagePath") ) {
					$err_msg = "ERR/Resize failed; cannot find NewImagePath '$NewImagePath'"
					if ($script:post_to_BI_logfile_flg -and $script:post_all_events) {
						PostCustomBiLogEntry `
						-MsgStr "$err_msg ($script:ScriptName)" `
						-Level 2 `
						-SilentFlg $SilentFlg
					}
					if ($script:FromConsole) {
						Write-Host $err_msg
					}
					msg.exe * /TIME:10 $err_msg
					
					# Release resource to avoid memory leak
					$img.Dispose()  
					
					return @($oldSizeKb, 0, $oldWH, "")
					
				} else {
					$newSizeKb = [math]::Round((Get-Item -Path $NewImagePath).Length/1kB, 1)
					$newWH = "$newWidth x $newHeight"
				}
				
				# Show progress
				if ($script:FromConsole) {
					Write-Host "--- result #$i height/width/size(KB) = $newHeight/$newWidth/$newSizeKb"
				}

				if ($newSizeKb -lt [Int]$TargetKb) {
					if ($script:FromConsole) {
						Write-Host "Achieved target after $i iterations"
					}
					break				
				} else {
					if ($q -le $step) {
						if ($script:FromConsole) {
							Write-Host ("Unable to achieve target after $i iterations; using final try results")
						}
					}
				}	
				$i++
			}
		}
			
		if ($TargetKbMethod -eq 'Hybrid') {
			$i = 1
			$step = $TargetKbWxHDelta
			for ($x=(100-$step); $x -gt 0; $x=$x-$step) {
				
				# Show progress
				if ($script:FromConsole) {
					Write-Host ("Trying hybrid reduction: size = -" + (100-$x) + "%; quality = -" + ($i*$TargetKbQualityDelta) + "%")
				}
				
				$j = $x/100
				$CanvasHeight = $oldHeight*$j
				$CanvasWidth  = $oldWidth*$j
				$q = $Quality - $i*$TargetKbQualityDelta

				# Encoder parameter for image quality
				$ImageEncoder = [System.Drawing.Imaging.Encoder]::Quality
				$encoderParams = New-Object System.Drawing.Imaging.EncoderParameters(1)
				$encoderParams.Param[0] = New-Object System.Drawing.Imaging.EncoderParameter($ImageEncoder, $q)
			 
				# Get codec
				$Codec = [System.Drawing.Imaging.ImageCodecInfo]::GetImageEncoders() | Where {$_.MimeType -eq 'image/jpeg'}
			 
				# Compute the final ratio to use
				$ratioX = $CanvasWidth / $img.Width;
				$ratioY = $CanvasHeight / $img.Height;
			 
				$ratio = $ratioY  # Default
				if ($ratioX -le $ratioY) {
					$ratio = $ratioX
				}
			 
				$newWidth = [Int] ($img.Width * $ratio)
				$newHeight = [Int] ($img.Height * $ratio)

				$bmpResized = New-Object System.Drawing.Bitmap($newWidth, $newHeight)
				#$bmpResized = New-Object System.Drawing.Bitmap -ArgumentList $newWidth, $newHeight
				$graph = [System.Drawing.Graphics]::FromImage($bmpResized)
				$graph.InterpolationMode = [System.Drawing.Drawing2D.InterpolationMode]::HighQualityBicubic
			 
				$graph.Clear([System.Drawing.Color]::White)
				$graph.DrawImage($img, 0, 0, $newWidth, $newHeight)
			 
				# Save to file
				$bmpResized.Save($OutputLocation, $Codec, $($encoderParams))
				$bmpResized.Dispose()
			
				if (-not(Test-Path -LiteralPath "$NewImagePath") ) {
					$err_msg = "ERR/Resize failed; cannot find NewImagePath '$NewImagePath'"
					if ($script:post_to_BI_logfile_flg -and $script:post_all_events) {
						PostCustomBiLogEntry `
						-MsgStr "$err_msg ($script:ScriptName)" `
						-Level 2 `
						-SilentFlg $SilentFlg
					}
					if ($script:FromConsole) {
						Write-Host $err_msg
					}
					msg.exe * /TIME:10 $err_msg
					
					# Release resource to avoid memory leak
					$img.Dispose()  
					
					return @($oldSizeKb, 0, $oldWH, "")
					
				} else {
					$newSizeKb = [math]::Round((Get-Item -Path $NewImagePath).Length/1kB, 1)
					$newWH = "$newWidth x $newHeight"
				}
				
				# Show progress
				if ($script:FromConsole) {
					Write-Host "--- result #$i height/width/size(KB) = $newHeight/$newWidth/$newSizeKb"
				}

				if ($newSizeKb -lt [Int]$TargetKb) {
					if ($script:FromConsole) {
						Write-Host "Achieved target after $i iterations"
					}
					break				
				} else {
					if (($x -le $step) -or ($q -le $TargetKbQualityDelta)) {
						if ($script:FromConsole) {
							Write-Host ("Unable to achieve target after $i iterations; using final try results")
						}
					}
					# Look ahead (necessary b/c $q is not managed by for:next loop)
					if (($q - $TargetKbQualityDelta) -lt $TargetKbQualityDelta) {
						break
					}
				}	
				$i++
			}
		}
		
		if (@('WxH', 'Quality', 'Hybrid') -contains $TargetKbMethod ) {
			if ($script:FromConsole) {
				Write-Host ""
			}
		}

		if ($script:FromConsole) {
			Write-Host "File resize results:"
			Write-Host "--------------------"
			Write-Host "SourceImagePath = '$SourceFilespec'"
			Write-Host "NewImagePath    = '$NewImagePath'"
			Write-Host "Original height/width/size(KB) = $oldHeight/$oldWidth/$oldSizeKb"
			Write-Host "Resized height/width/size(KB)  = $newHeight/$newWidth/$newSizeKb"
		}
				
		# Release resource to avoid memory leak
		$img.Dispose()  
		
		return @($oldSizeKb, $newSizeKb.ToString("#.0"), $oldWH, $newWH)

	} else {
		
		if (($Quality -eq 100) -and ($Width -eq 0)) {

			# SPECIAL CASE - USE THE SOURCE IMAGE AS-IS

			$newHeight = $oldHeight
			$newWidth  = $oldWidth
			$newSizeKb = $oldSizeKb
			
			# Copy the original image file 'as-is'
			if (Test-Path -LiteralPath "$NewImagePath") {
				Remove-Item -LiteralPath "$NewImagePath" -Force
			}
			Copy-Item -LiteralPath "$SourceFilespec" -Destination "$NewImagePath"			
			# ... and give it a current timestamp
			#     WHY? w/o a new timestamp, it may be deleted by 'PurgeOldFiles()'
			$file = Get-Item "$NewImagePath"
			$file.LastWriteTime = (Get-Date)	

		} else {

			# NORMAL CASE - RESIZE THE SOURCE IMAGE

			$CanvasWidth  = $Width
			$CanvasHeight = ($Width / $oldWidth) * $oldHeight
		 
			# Encoder parameter for image quality
			$ImageEncoder = [System.Drawing.Imaging.Encoder]::Quality
			$encoderParams = New-Object System.Drawing.Imaging.EncoderParameters(1)
			$encoderParams.Param[0] = New-Object System.Drawing.Imaging.EncoderParameter($ImageEncoder, $Quality)
		 
			# Get codec
			$Codec = [System.Drawing.Imaging.ImageCodecInfo]::GetImageEncoders() | Where {$_.MimeType -eq 'image/jpeg'}
		 
			# Compute the final ratio to use
			$ratioX = $CanvasWidth / $img.Width;
			$newWidth  = [Int] ($img.Width * $ratioX)
			$newHeight = [Int] ($img.Height * $ratioX)
			
			#Write-Host "newWidth = $newWidth"
			#Write-Host "newHeight = $newHeight"

			$bmpResized = New-Object System.Drawing.Bitmap($newWidth, $newHeight)
			#$bmpResized = New-Object System.Drawing.Bitmap -ArgumentList $newWidth, $newHeight
			$graph = [System.Drawing.Graphics]::FromImage($bmpResized)
			$graph.InterpolationMode = [System.Drawing.Drawing2D.InterpolationMode]::HighQualityBicubic
		 
			$graph.Clear([System.Drawing.Color]::White)
			$graph.DrawImage($img, 0, 0, $newWidth, $newHeight)
		 
			# Save to file
			$bmpResized.Save($OutputLocation, $Codec, $($encoderParams))
			$bmpResized.Dispose()
			$img.Dispose()
		
			if (-not(Test-Path -LiteralPath "$NewImagePath") ) {
				$err_msg = "ERR/Resize failed; cannot find NewImagePath '$NewImagePath'"
				if ($script:post_to_BI_logfile_flg -and $script:post_all_events) {
					PostCustomBiLogEntry `
					-MsgStr "$err_msg ($script:ScriptName)" `
					-Level 2 `
					-SilentFlg $SilentFlg
				}
				if ($script:FromConsole) {
					Write-Host $err_msg
				}
				msg.exe * /TIME:10 $err_msg
				
				# Release resource to avoid memory leak
				#$img.Dispose()  

				return @($oldSizeKb, $null)
			} else {
				$newSizeKb = [math]::Round((Get-Item -Path $NewImagePath).Length/1kB, 1).ToString("#.0")
			}
		}
		if ($script:FromConsole) {
			Write-Host "File resize results:"
			Write-Host "-------------------"
			Write-Host "SourceImagePath = '$SourceFilespec'"
			Write-Host "NewImagePath    = '$NewImagePath'"
			Write-Host "Original height/width/size(KB) = $oldHeight/$oldWidth/$oldSizeKb"
			Write-Host "Resized height/width/size(KB)  = $newHeight/$newWidth/$newSizeKb"
		}
	}
	
	$oldWH = "$oldWidth x $oldHeight"
	$newWH = "$newWidth x $newHeight"
	
	# For debugging
	#Write-Host "oldWH = $oldWH"
	#Write-Host "newWH = $newWH"
	
	# Release resource to avoid memory leak
	#$img.Dispose()
	
	return @($oldSizeKb, $newSizeKb, $oldWH, $newWH)
}

function SendPushover() {
# PURPOSE: Send custom notification via the Pushover API
# ARGS:    $MsgStr        = (string) pushover notification message; may contain HTML
#          $MsgTitle      = (string) pushover notification title
#          $PriorityLevel = (integer) -2|-1|0|1|2; -2=lowest, -1=low(silent), 0=normal, 1=high-priority, 2=emergency
#          $AttachPath    = (string) full filespec of resized alert image, or $null
#          $Camera        = (string) Blue Iris camera shortname (replaces '&CAM' in user-setting $po_tags)
# RETURNS: (string) Pushover API response, or 'ERR/' message
# GLOBALS: Uses vars: $script:po_app_token, $script:po_user_token, $script:retry, $script:po_expire, 
#                     $script:po_sound, $script:po_device, $script:po_tags, $script:server_username
# ATTRIBUTIONS:
#    handling attachments: https://tinyurl.com/mrxp7pn3 (PoshOver Powershell Package)
#
# V4 (9/8/2024) eliminated arg -QuietFlg; modified usage/support of global vars
# By jaydeel @ ipcamtalk.com
	
    param(
		[String]$MsgStr,      # [OPTIONAL] default is 'test'
		[String]$MsgTitle,    # [OPTIONAL] default is the App name associaled with your Pushover App token
		[Int]$PriorityLevel,  # [OPTIONAL] default is 0
		[String]$AttachPath,  # [OPTIONAL] default is null
		[String]$Camera      # [OPTIONAL] default is ""
	)
	
	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	# Write-Host "Entered function: $fn_name()"
	# Write-Host "$fn_name() MsgStr = '$MsgStr'"
	# Write-Host "$fn_name() MsgTitle = '$MsgTitle'"
	# Write-Host "$fn_name() PriorityLevel = PriorityLevel"
	# Write-Host "$fn_name() AttachPath = '$AttachPath'"	
	# Write-Host "$fn_name() Camera = '$Camera'"	
	# Write-Host ("$fn_name() Sound = '" + $script:po_sound)
	# Write-Host ("$fn_name() Devices = '" + $script:po_device)
	# NOTE REGARDING MSG.EXE COMMANDS
	# ... You may need to replace 'msg.exe *' with 'msg.exe /server:computername username
	#msg.exe $script:server_username /TIME:20 "Pushover notification args...`npriority_level = $PriorityLevel`nAttachPath = $AttachPath`nMsgStr = $MsgStr"

	# HANDLE OPTIONAL COMMANDLINE ARGS
	if (!$MsgStr) {
		$MsgStr = "Alert w/ GIF attachment demo"
	}
	if (!$Camera) {
		$cam = $Camera
	} else {
		$cam = ''
	}

	$rtn_str = $null
	
	#Prep the attachment
	if ($AttachPath) {
		#Get-Item -Path $AttachPath
		$attach_obj = (Get-Item -Path $AttachPath)
		$attach_bytes = [System.IO.File]::ReadAllBytes($AttachPath)
	}

	# For debugging
	#Write-Host ("$fn_name() attach_obj = " + $attach_obj)
	#Write-Host ("$fn_name() attach_obj.Name = " + $attach_obj.Name)
	#Write-Host ("$fn_name() po_tags = '" + $script:po_tags + "'")
	#Write-Host ("$fn_name() Camera = " + $Camera)
	
	# PREP CORE PARAMS
	# ... will be modified if sending an attachment
	$uri = 'https://api.pushover.net/1/messages.json'
	if ($script:po_tags) {
		$po_tags = ($script:po_tags).Replace("&CAM", $cam)
	} else {
		$po_tags = ''		
	}
	$parameters = @{
		token      = $script:po_app_token
		user       = $script:po_user_token
		priority   = $PriorityLevel
		retry      = $script:po_retry
		expire     = $script:po_expire
		tags       = $po_tags
		sound      = $script:po_sound
		device     = $script:po_device
		html       = 1
		title      = $MsgTitle
		message    = $MsgStr
	}
	
	# For debugging
	# Write-Host ("parameters.Keys = " + $parameters.Keys)
	# Write-Host "`nPushover parameters..."
	<# [string[]]$arr = $parameters | Out-String -stream
	foreach ($item in $arr) {
		if ($item) {
			Write-Host $item
		}
	} #>
	
	# EXECUTE
	if (!$AttachPath -or ($attach_bytes.Length -eq 0)) {
		# SEND BASIC NOTIFICATION

		# ... NOTE: When running in the background, both Invoke-RestMethod & Invoke-WebRequest
		#	  do not execute; it doesn't produce anything, no error, warning, or output.
		#     REF: https://stackoverflow.com/a/50274056 looked like a possible solution
		#     however, since PS v6.0 switch -UseBasicParsing has been deprecated, and 
		#     all web requests now use basic parsing only.
		# ... attachmnets
		#     https://psg-prod-centralus.cloudapp.net/packages/Poshover/0.1.450600-alpha/Content/Private%5CSend-MessageWithAttachment.ps1
		$ApiResponse = $parameters | Invoke-RestMethod -Uri $uri -Method Post
		# Alternative syntaxes
		#$parameters | Invoke-RestMethod -Uri $uri -Method Post | Out-Null
		#$ApiResponse = Invoke-RestMethod -Uri $uri -Method Post -Form $parameters
		
		$response_str = ($ApiResponse | ConvertTo-Json) | ConvertFrom-Json
				
		# For debugging
		#Write-Host "$fn_name() w/o attachment ApiResponse = $response_str"

	} else {	
		# SEND NOTIFICATION WITH ATTACHMENT
		
		# Notes:
		# Modified code from:  https://tinyurl.com/mrxp7pn3
		# More refs: https://gist.github.com/markekraus/a0135d0a5f52b609eda513df4db71cab

		try {

            Add-Type -AssemblyName System.Net.Http
			# NOTE: Originally, the $client= line below was not working when this script was 
			#        executed in the background via a Blue Iris 'run a program' action
			#        SOLUTION:  https://stackoverflow.com/a/70535083
			$client = [system.net.http.httpclient]::new()

			try {
				$content = [system.net.http.multipartformdatacontent]::new()
				foreach ($key in $parameters.Keys) {
					$textContent = [system.net.http.stringcontent]::new($parameters.$key)
					$content.Add($textContent, $key)
				}
				
				# For debugging
				#Write-Host ("content = " + $content | ConvertTo-Json)
				
				$jpegContent = [system.net.http.bytearraycontent]::new($attach_bytes)
				$jpegContent.Headers.ContentType = [system.net.http.headers.mediatypeheadervalue]::new('image/jpeg')
				$jpegContent.Headers.ContentDisposition = [system.net.http.headers.contentdispositionheadervalue]::new('form-data')
				$jpegContent.Headers.ContentDisposition.Name = 'attachment'
				$jpegContent.Headers.ContentDisposition.FileName = $AttachPath
				$content.Add($jpegContent)

				# Not needed.
				# REF: https://stackoverflow.com/questions/43014894/httpclient-and-stream-issues
				#$content.LoadIntoBufferAsync().GetAwaiter().GetResult()

				# Modify the message body
				Write-Verbose "Message body:`r`n$($content.ReadAsStringAsync().Result.Substring(0, 2000).Replace($parameters.token, "********").Replace($parameters.user, "********"))"

				# For debugging
				#msg.exe $script:server_username /TIME:10 "executing pushover api cmd w/ attachment..."

				# DO IT!
				$ApiResponse = $client.PostAsync($uri, $content).Result
				
				# For debugging
				#Write-Host ("$fn_name() w/ attachment ApiResponse = " + ($ApiResponse | ConvertTo-Json -Depth 32))
				
				try {
					$response_str = $ApiResponse.Content.ReadAsStringAsync().Result | ConvertFrom-Json
					
					# For debugging
					#Write-Host "$fn_name() ApiResponse.Content = $response_str"
				}
				catch {
					$response_str = "WARNING: Unexpected Pushover API response; attachment may be too large"
				}
				
				# For debugging
				#Write-Host "$fn_name() w attachment ApiResponse = $response_str"
				
			}
			catch {
				$response_str = "ERR/unable to send Pushover notification"				
			}
			finally {
				$content.Dispose()  # Release resources
			}
		}
		finally {
			$client.Dispose()  # Release resources
		}
	}

	# For testing
	
	if ($response_str -match 'invalid') {
		$rtn_str = "ERR/Pushover API returned 'invalid'"
	} elseif (!($response_str -match 'status=1')) {
		$rtn_str = "ERR/failed to send Pushover notification"				
	} else {
		# SUCCESS!!
		$rtn_str = ("Pushover notification sent`nAPI response = " + $response_str)
	}
	
	return $rtn_str
}

function LogScriptResultsToCsv() {
# PURPOSE:  Log performance data to CSV file.
# ARGS:     see param lines below
# RETURNS:  (string) 'success' or 'ERR/unable to append row to CSV file'
# NOTES:    1. CSV file is created automatically if it does not already exist. 
#           2. CSV file's header row is generated from the object keys in variable $new_row_obj.
	
    param(
		[String]$csv_filespec, 
		[String]$img_filename, 
		[String]$timestamp, 
		[double]$before_kb, 
		[double]$after_kb, 
		[String]$before_wh, 
		[String]$after_wh, 
		[double]$total_ms, 
		[double]$resize_ms, 
		[double]$pusho_ms
	)
	
	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	#Write-Host "Entered function = $fn_name()"
	
	$rtn_str = 'success'
		
	# Build the new row (object keys are the header column labels)
	$new_row_obj = [PSCustomObject]@{
		'Timestamp'    = $timestamp
		'SizeBeforeKb' = $before_kb
		'SizeAfterKb'  = $after_kb
		'WxHBefore'    = $before_wh
		'WxHAfter'     = $after_wh
		'TotalMsec'    = $total_ms
		'ResizeMsec'   = $resize_ms
		'PushoMsec'    = $pusho_ms
		'ImageFile'    = $img_filename
	}
	
	# For debugging
	#msg.exe * /TIME:10 "new_row_obj... `n$new_row_obj"

	# Append new row to CSV file, creating the file if necessary
	if ($PSVersionTable.PSVersion.Major -ge 7) {
		# REFS: https://tinyurl.com/yvy4mjy6
		#       https://stackoverflow.com/a/70717331
		#       https://stackoverflow.com/a/60681762
		try {
			$new_row_obj | Export-Csv -LiteralPath $csvfilespec -Append -Force -UseQuotes AsNeeded
		} catch {
			$err_msg = "ERR/unable to append row to '$csvfilespec'"
			return $err_msg
		}
	} else {
		try {
			$new_row_obj | Export-Csv -LiteralPath $csvfilespec -Append -Force -NoTypeInformation
		} catch {
			$err_msg = "ERR/unable to append row to '$csvfilespec'"
			return $err_msg
		}
	}
	
	return $rtn_str
}

function PurgeOldFiles() {
# PURPOSE:  Maintain the resized images folder; delete resized image files older than N hours 
# ARGS:     see param lines below
# RETURNS:  nothing
# ATTRIBUTIONS:  https://stackoverflow.com/a/46086926
	
    param(
		[String]$Pattern,  # folder location and filter for image files; a filespec like 'D:\myfolder\*.jpg'
		[Int]$Hours        # maximum age for retained files; use 24 for 1 day, 168 for 1 week... 
	)
	
	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	#Write-Host "Entered function = $fn_name()"

	# GET FILES MATCHING AGE REQUIREMENT
	$cur_list = get-childitem -Path $Pattern | Select Name
	$limit = (Get-Date).AddHours(-$Hours)
	$purge_obj = Get-ChildItem -Path $Pattern | Where-Object { !$_.PSIsContainer -and $_.CreationTime -lt $limit } | Select Name
	$purge_list = $purge_obj.Name
	
	if ($cur_list) {
		$cur_list_cnt = $cur_list.length
	} else {
		$cur_list_cnt = 0
	}

	if ($purge_list) {
		$purge_list_cnt = $purge_list.length
	} else {
		$purge_list_cnt = 0
	}

	# REMOVE FILES MATCHING AGE REQUIREMENT
	if ($purge_list_cnt -gt 0) {
		if ($Hours) {
			$null = Get-ChildItem -Path $Pattern | 
				Where-Object { !$_.PSIsContainer -and $_.CreationTime -lt $limit } | 
				Remove-Item -Force
		}
	}
	
	return  @($cur_list_cnt, $purge_list.Count)
}

# Start overall performance timer
$timer_total = [Diagnostics.Stopwatch]::StartNew()

if ($Help -or $cmdline_args_passed.length -eq 0) {
	if ($FromConsole) {
		Write-Host "======================================================================================="
		Write-Host "Pushover Notification with Resized Image Attachment"
		Write-Host "---------------------------------------------------------------------------------------"
		Write-Host "Usage: $ScriptName -Msg '' -AttachFile '' [optional arguments]"
		Write-Host "Arguments: "
		Write-Host "  -Msg ... message for Pushover notification; may contain HTML"
		Write-Host "  -AttachFile ... full filespec for Pushover notification attachment to be resized [1]"
		Write-Host "  -Quality  ... attachment JPG image quality; 1 to 100 [2,5]"
		Write-Host "  -Width    ... attachment JPG image width (pixels); aspect ratio is preserved [2,5]"
		Write-Host "  -TargetKB ... target size of resized attachment; the new file will be <= this value"
		Write-Host "  -Method   ... custom method for arg -TargetKB: 'WxH' (default), 'Quality' or 'Hybrid'"
		Write-Host "  -Priority ... custom priority for Pushover notification, integer, see Notes [3,5]"
		Write-Host "  -Sound    ... custom sound for Pushover notification [4,5]"
		Write-Host "  -Devices  ... custom device list for Pushover notification; comma-delimited [5]"
		Write-Host "  -Demo ... skips notification & BI logfile posting (useful when tuning TargetKB)"
		Write-Host "Notes:"
		Write-Host "[1] When using script in a Blue Iris 'On alert' Action, use '[path/]&ALERT_PATH'"
		Write-Host "[2] Using '-Quality 100 -Width 0' sends AttachFile without resizing"
		Write-Host "[3] -2=lowest, -1=low(silent), 0=normal, 1=high-priority, 2=emergency"
		Write-Host "[4] This sound usually plays ONLY if the Pushover app is in the device's background"
		Write-Host "[5] These arguments override the default values in the user-setting file"
		Write-Host "---------------------------------------------------------------------------------------"
		Write-Host ""
	}
	exit
}

# HANDLE COMMANDLINE ARGS
if (1 -eq 1) {  # Use to collapse this code section while editing
	
	if ($Msg -and $AttachFile) {
		$po_msg = $Msg
	} else {
		$err_msg = "ERR/expecting 2 required arguments (-Msg and -AttachFile)"
		if ($post_to_BI_logfile_flg) {
			PostCustomBiLogEntry `
			-MsgStr "$err_msg ($ScriptName)" `
			-Level 2 `
			-SilentFlg ($bg_msgs_flg -eq $false)
		}
		if ($FromConsole) {
			return "$err_msg"
		} else {
			msg.exe * /TIME:10 $err_msg
		}	
	}
	
	if ($Priority -or ($Priority -eq 0)) {
		if ($Priority -ne 0) {
			if (($Priority -ge -2) -and ($Priority -le 2)) {
				$po_priority = $Priority
			}
		}
	}
	
	$attach_path = $null
	$variable_test = Get-Variable -Name AttachFile -Scope Local -ErrorAction SilentlyContinue
	if ($null -ne $variable_test) {
		$attach_path = $AttachFile
	}
	#Write-Host "attach_path = '$attach_path'"
	
	#$img_quality = 85  # user-settings default 
	if ($Quality) {
		if ($Quality -ge 1 -and $Quality -le 100) {
			$img_quality = $Quality
		}
	}
	
	#$img_width = 500  # user-settings default
	if ($Width) {
		$img_width = $Width
	}
	
	#$img_kilobytes = 0  # user-settings default
	if ($TargetKB -or $TargetKB -eq 0) {
		$img_kilobytes = $TargetKB
	}

	if ($Method) {
		if (@('WxH','Quality','Hybrid') -notcontains $Method) {
			$err_msg = "ERR/expecting -Method to be 'WxH', 'Quality', or 'Hybrid')"
			if ($FromConsole) {
				return "$err_msg"
			} else {
				msg.exe * /TIME:10 $err_msg
			}	
		} else {
			$targetkb_method = $Method
		}
	}

	
	if ($Sound) {
		$script:po_sound = $Sound
	}
	
	if ($Devices) {
		$script:po_device = $Devices
	}
	
	if ($Quiet) {
		$bg_msgs_flg = $false
	}
}

# ABORT IF PATHS NOT FOUND
if (1 -eq 1) {  # Use to collapse this code section while editing

	if (-not(Test-Path -LiteralPath "$img_path") ) {
		$err_msg = "ERR/destination path does not exist"
		<# if ($post_to_BI_logfile_flg) {
			PostCustomBiLogEntry `
			-MsgStr "$err_msg ($ScriptName)" `
			-Level 2 `
			-SilentFlg ($bg_msgs_flg -eq $false)
		} #>
		if ($FromConsole) {
			return $err_msg
		} else {
			msg.exe * /TIME:10 "$err_msg`n`n(via script: '$ScriptName')"
		}	
	}

	if ($attach_path) {
		if (-not(Test-Path -LiteralPath "$attach_path") ) {
			$err_msg = "ERR/alert image file not found (expecting the full filespec); '$attach_path'"
			if ($post_to_BI_logfile_flg) {
				PostCustomBiLogEntry `
				-MsgStr "$err_msg ($ScriptName)" `
				-Level 2 `
				-SilentFlg ($bg_msgs_flg -eq $false)
			}
			if ($FromConsole) {
				return $err_msg
			} else {
				msg.exe * /TIME:10 "$err_msg`n`n(via script: '$ScriptName')"
			}	
		}
	}

}

# RESIZE ATTACHMENT
if ($attach_path) {

	# Create resized image's full filespec
	$full_filespec = (Resolve-Path $attach_path).Path
	$dot_idx = $full_filespec.LastIndexOf(".")
	if ($img_name_suffix) {
		$resized_image_path = $full_filespec.Substring(0,$dot_idx) + "_" + $img_name_suffix + $full_filespec.Substring($dot_idx,$full_filespec.Length - $dot_idx)
	} else {
		#$resized_image_path = $full_filespec.Substring(0,$dot_idx) + $full_filespec.Substring($dot_idx,$full_filespec.Length - $dot_idx)
		$resized_image_path = $full_filespec
	}

	# Change destination path, if requested
	$basename = Split-Path $attach_path -leaf
	$source_path = $full_filespec.Replace($basename, "")
	$resized_image_path = $resized_image_path.Replace($source_path, $img_path)

	# For debugging
	#Write-Host "source image full_filespec = '$full_filespec'"
	#Write-Host "source image filename = '$basename'"
	#Write-Host "source image source_path = '$source_path'"
	#Write-Host "resized_image_path = '$resized_image_path'"

	# Start resize performance timer
	$timer_resize = [Diagnostics.Stopwatch]::StartNew()

	# Resize attachment
	$resize_ary = `
		ResizeImage `
		-SourceFilespec $attach_path `
		-NewImagePath $resized_image_path `
		-Quality $img_quality `
		-Width $img_width `
		-TargetKb $img_kilobytes `
		-TargetKbMethod $targetkb_method `
		-TargetKbWxHDelta $targetkb_wxh_step `
		-TargetKbQualityDelta $targetkb_quality_step `
		-SilentFlg ($bg_msgs_flg -eq $false)

	# Stop resize performance timer
	$timer_resize.stop()
	$resize_ms = [math]::Round($timer_resize.elapsed.totalmilliseconds,1).ToString("0.0")

	if ($FromConsole) {
		Write-Host ("File resizing time (msec) = " + $resize_ms)
	}
}

# Start pushover performance timer
$timer_pusho = [Diagnostics.Stopwatch]::StartNew()

# SEND NOTIFICATION
if (!$Demo) {
	if (!($cmdline_args_passed -contains "Title")) {
		$po_title = "via $script:ScriptName"
	} else {
		$po_title = $Title
	}
	if ($attach_path) {
		if ($resize_ary[1] -eq 0) {
			# Try sending the original, unsized attachment
			$result_str = `
				SendPushover `
				-MsgTitle $po_title `
				-MsgStr $po_msg `
				-AttachPath $attach_path `
				-PriorityLevel $po_priority
				
			if ($FromConsole) {
				Write-Host "HEADS-UP: Requested attachment not resized; tried sending notification with unresized file"
			}
		} else {
			$result_str = `
				SendPushover `
				-MsgTitle $po_title `
				-MsgStr $po_msg `
				-AttachPath $resized_image_path `
				-PriorityLevel $po_priority
		}
	} else {
		$result_str = `
			SendPushover `
			-MsgTitle $po_title `
			-MsgStr $po_msg `
			-AttachPath $null `
			-PriorityLevel $po_priority
	}
	
	if ($FromConsole) {
		Write-Host ""
		Write-Host "Pushover response:"
		Write-Host "------------------"
		Write-Host $result_str
	}
}

# Stop pushover performance timer
$timer_pusho.stop()
$pusho_ms = [math]::Round($timer_pusho.elapsed.totalmilliseconds,1).ToString("0.0")

if (!$Demo) {
	if ($FromConsole) {
		Write-Host ("Pushover execution time (msec) = " + $pusho_ms)
	}
}

# REMOVE OLD REDUCED IMAGE FILES
if (!$Demo) {
	$purge_filespec = $img_path + "*.jpg"
	#Write-Host "purge_filespec = '$purge_filespec'"
	$arr = `
		PurgeOldFiles `
		-Pattern $purge_filespec `
		-Hours $img_lifetime_hours
	
	if ($FromConsole) {
		Write-Host ""
		Write-Host "File pruning:"
		Write-Host "-------------"
		Write-Host ("Count of archived JPG files matching '$purge_filespec' = " + $arr[0])
		if ($arr[1] -gt 0) {			
			Write-Host ("Count of archived JPG files deleted (age >= $img_lifetime_hours hrs) = " + $arr[1])
		} else {
			Write-Host "No files pruned"
		}
	}
}

# Stop overall performance timer
$timer_total.stop()
$total_ms  = [math]::Round($timer_total.elapsed.totalmilliseconds,1).ToString("0.0")

# LOG SCRIPT PERFORMANCE INFO
$csvfilespec = $img_path + $ScriptName + ".csv"
$timestamp   = Get-Date -Format "yyyy/MM/dd_HH:mm:ss"
$total_ms    = [decimal]$total_ms 
$pusho_ms    = [decimal]$pusho_ms 
if ($attach_path) {
	$size_before = [decimal]$resize_ary[0]
	$size_after  = [decimal]$resize_ary[1]
	$wxh_before  = [String]$resize_ary[2]
	$wxh_after   = [String]$resize_ary[3]
	$resize_ms   = [decimal]$resize_ms 
	$result_str = `
		LogScriptResultsToCsv `
		$csvfilespec $attach_path $timestamp $size_before $size_after $wxh_before $wxh_after $total_ms $resize_ms $pusho_ms
} else {
	$size_before = 0
	$size_after  = 0
	$wxh_before  = ""
	$wxh_after   = ""
	$resize_ms   = 0 
	$result_str = `
		LogScriptResultsToCsv `
		$csvfilespec 'NA' $timestamp $size_before $size_after $wxh_before $wxh_after $total_ms $resize_ms $pusho_ms
}

# Performance notice
if ($FromConsole) {
	$perf_msg = "Script execution time (msec) = " + [math]::Round($timer_total.elapsed.totalmilliseconds,1)
	Write-Host ""
	Write-Host "Performance summary:"
	Write-Host "--------------------"
	Write-Host $perf_msg
}

$wxh_before     = $wxh_before.Replace(" x ", "x")
$wxh_after      = $wxh_after.Replace(" x ", "x")
$disp_total_ms  = [math]::Round($total_ms,0)
$disp_resize_ms = [math]::Round($resize_ms,0)
$disp_pusho_ms  = [math]::Round($pusho_ms,0)

# SHOW OPTIONAL MESSAGE WHEN RUNNING FROM BACKGROUND
if (!$FromConsole) {
	if ($bg_msgs_flg) {
		$msg_sec = 30
		$bg_msg =  $attach_path + 
			"`n`nWxH before/after  = $wxh_before / $wxh_after" +
			"`n`nSize before/after (KB) ... $size_before / $size_after" + 
			"`n`nexecution times (resize,pusho,total) ... $disp_resize_ms, $disp_pusho_ms, $disp_total_ms msec" + 
			"`n`n(this msg auto-closes in $msg_sec seconds)"
		msg.exe * /TIME:$msg_sec $bg_msg
	}
}

# POST MESSAGE TO BLUE IRIS LOGFILE
# KEY POINT: the message string maximum length is 128 characters (BI HTTP interface command '/admin?log=message&level=x')
if (!$Demo) {
	if ($post_to_BI_logfile_flg) {
		$log_msg = "pushover resize [$wxh_before, $wxh_after; $size_before KB, $size_after KB; $disp_resize_ms msec, $disp_pusho_ms msec, $disp_total_ms msec] file=$attach_path"		
		$result_str = `
			PostCustomBiLogEntry `
			-MsgStr "$log_msg" `
			-Level 0 `
			-SilentFlg $true

		if ($FromConsole) {
			Write-Host ""
			Write-Host "Blue Iris logfile:"
			Write-Host "------------------"
			Write-Host $result_str
		}

	}
}

if ($FromConsole) {
	Write-Host ""
}

#Stop-Transcript